--SkyRide Universal Flight Core
--Made by: Kanderman
--Do not redistribute

local M = {}

M.active = false
M.targetAlt = 0
M.thrust = 0
M.yawthr = 0
M.altMode = false
M.targetAltAdj = 0

M.stabs = {}
M.particleEmitters = {}
M.totalMass = 0

local stabs = {}

local AltAuth = 0

local PitchAuth = 0
local RollAuth = 0
local YawAuth = 0

local tTorq = vec3(0,0,0)

local lastYaw = 0

local flight_noise

local transform_noise

local timer = 0

local ls = false

--PID controllers
local aPID = newPIDParallel(0.6, 0.0125, 0.3, -1, 4) --Altitude

local pPID = newPIDParallel(12, 0, 2, -6, 6) --Pitch
local rPID = newPIDParallel(6, 0, 1, -6, 6) --Roll

local min = math.min
local max = math.max
local clamp = function(v, mi, ma) return max(min(v, ma), mi)end
local clamp2 = function(v, c) return clamp(v, -c, c) end

local function changeMode()
	if electrics.values.SkyRide_mode == 1 then
		electrics.values.SkyRide_mode = 0
		extensions.skyride_skydros.toggleModeActive(false)
		M.active = false
		controller.mainController.setEngineIgnition(true)
		controller.mainController.setStarter(true)
	elseif electrics.values.SkyRide_mode == 0 and electrics.values.ignition ~= 0 then
		if v.data.energyStorage then
			local fuel = 0
			for _, tank in pairs(v.data.energyStorage) do
				if tank.type == "fuelTank" then
					fuel = fuel + energyStorage.getStorage(tank.name).remainingVolume
				end
			end
			if fuel <= 0 then
				guihooks.message("No fuel, refill your tank to fly", 3, "SkyRide")
				return
			end
		end

		electrics.values.SkyRide_mode = 1
		extensions.skyride_skydros.toggleModeActive(true)
		timer = 0.25
		controller.mainController.setEngineIgnition(false)
		if M.altMode then
			M.targetAlt = obj:castRayStatic(obj:getFrontPosition(), vec3(0,0,-1), 1000) + 3.5
			guihooks.message("Terrain follow mode: ACTIVE", 3, "SkyRide")
		else
			M.targetAlt = electrics.values.altitude + 3
			guihooks.message("Terrain follow mode: INACTIVE", 3, "SkyRide")
		end
	else
		guihooks.message("Turn on the ignition to fly", 3, "SkyRide")
		return
	end
end

local function altChange(val)
	M.targetAltAdj = val*5
end

local function hoverMode()
	if M.altMode then
		M.altMode = false
		M.targetAlt = electrics.values.altitude
		guihooks.message("Terrain follow mode: INACTIVE", 3, "SkyRide")
	else
		M.altMode = true
		M.targetAlt = obj:castRayStatic(obj:getFrontPosition(), vec3(0,0,-1), 1000)
		guihooks.message("Terrain follow mode: ACTIVE", 3, "SkyRide")
	end
end

local function update(dt)
	if ls ~= (electrics.values.SkyRide_mode == 1) then
		if electrics.values.SkyRide_mode == 1 then
			obj:playSFX(flight_noise)
		else
			obj:stopSFX(flight_noise)
		end
		obj:playSFX(transform_noise)
		ls = electrics.values.SkyRide_mode == 1
	end
	if timer > 0 then
		timer = timer - dt
		if timer <=0 and electrics.values.SkyRide_mode == 1 then
			M.active = true
		end
	end
	if M.active then
		local fuel = 0
		local tc = 0
		if v.data.energyStorage then
			for _, tank in pairs(v.data.energyStorage) do
				if tank.type == "fuelTank" then
					fuel = fuel + energyStorage.getStorage(tank.name).remainingVolume
					tc = tc + 1
				end
			end
			if fuel <= 0 then
				guihooks.message("No fuel, refill your tank to fly", 3, "SkyRide")
				electrics.values.SkyRide_mode = 0
				M.active = false
				return
			end
		end
		if electrics.values.rpm > 2 then
			controller.mainController.setEngineIgnition(false)
		end
		controller.mainController.setStarter(false)

		local vPos = obj:getPosition()
		local vVel = obj:getVelocity()
		local vDir = obj:getDirectionVector()
		local vDirR = obj:getDirectionVectorRight()
		local vPosF = obj:getFrontPosition()

		local roll, pitch, yaw = obj:getRollPitchYaw()

		local vVelL = vVel:length()
		local vVelX = vVel:dot(vDir)
		local vVelY = vVel:dot(vDirR)

		local yawVel = (yaw - lastYaw)/dt

		lastYaw = yaw

		local vCog = obj:calcCenterOfGravity(false)

		local vCogL = (vCog - vPos)
		local qx, qy, qz, qw = obj:getRotation()
		vCogL:setRotate(quat(qx, qy, qz, -qw))

		AltAuth = 0

		PitchAuth = 0
		RollAuth = 0
		YawAuth = 0

		tTorq:set(0,0,0)

		for k, stab in pairs(M.stabs) do
			local pos1 = (obj:getNodePositionRelative(stab.id1) - vCogL)
			local pos2 = (obj:getNodePositionRelative(stab.id2) - vCogL)
			local dir = (pos2 - pos1):normalized()
			local mags = pos2:cross(dir)
			local upd = dir:dot(vec3(0,0,-1))

			tTorq = tTorq + mags

			PitchAuth = PitchAuth + math.abs(mags.x)
			RollAuth = RollAuth + math.abs(mags.y)
			YawAuth = YawAuth + math.abs(mags.z)

			AltAuth = AltAuth + upd

			stab.pos = pos2
			stab.upd = upd
			stab.mags = mags

			stabs[k] = stab
		end

		PitchAuth = PitchAuth/#stabs
		RollAuth = RollAuth/#stabs
		YawAuth = YawAuth/#stabs

		AltAuth = AltAuth/#stabs

		local alt

		if M.altMode then
			local ddist = obj:castRayStatic(vPosF, vec3(0,0,-1), 5 + M.targetAlt*2)
			local fdist = obj:castRayStatic(vPosF + vec3(vVel.x, vVel.y, 0)*0.4,vec3(0,0,-1), 5 + M.targetAlt*2)
			alt = ddist - (fdist-ddist)*0.2
		else
			alt = vPos.z
		end

		local altAdj = ((2 + aPID:get(alt, M.targetAlt, dt)*2)/AltAuth)*#stabs

		local pitchadj = pPID:get(pitch, 0, dt)/PitchAuth
		local rolladj = rPID:get(roll, clamp2((vVelY/12)/(1+vVelL/30), 0.6), dt)/RollAuth
		local yawadj = ((clamp2(-electrics.values.steering_input*3, max(5/math.abs(vVelL)/30, 5)) - yawVel)*2)/YawAuth
		
		for _, stab in pairs(stabs) do
			local force = (altAdj*(AltAuth/stab.upd))*math.sqrt((tTorq.x/stab.mags.x)^2 + (tTorq.y/stab.mags.y)^2 + (tTorq.z/stab.mags.z)^2)
			force = force + pitchadj/(PitchAuth/stab.mags.x)
			force = force - rolladj/(RollAuth/stab.mags.y)
			force = force - yawadj/(YawAuth/stab.mags.z)

			obj:applyForce(stab.id2, stab.id1, clamp2(force, (60/#stabs)/(AltAuth/stab.upd))*M.totalMass)
		end

		local brake = electrics.values.brake_input

		if math.abs(vVelX) <= 0.2 then
			brake = brake*vVelX*0.1
		elseif vVelX <= 0 then
			brake = -brake
		end
		
		if electrics.values.gear == -1 then
			M.thrust = (-electrics.values.throttle_input/4 - brake*2)*10
		else
			M.thrust = (electrics.values.throttle_input - brake*2)*10
		end

		local thr = M.thrust/(1+math.abs(vVelX)/80) - (vVelX/35)

		for _, thruster in pairs(M.stabs) do
			obj:applyForceVector(thruster.id2, vDir*(thr*M.totalMass/#M.stabs))
		end

		local mul = 1

		if M.altMode and M.targetAlt < 8 then
			if M.targetAlt < 4 then
				mul = 0.25
			else
				mul = 0.5
			end
		end

		M.targetAlt = max(M.targetAlt + M.targetAltAdj*mul*dt, 0)

		--Fuel usage

		if v.data.energyStorage then
			local used = ((0.05+vVelL/200)*dt)/tc
			for _, tank in pairs(v.data.energyStorage) do
				if tank.type == "fuelTank" then
					local stor = energyStorage.getStorage(tank.name)
					stor:setRemainingVolume(stor.remainingVolume - used)
				end
			end
		end
	end
end

local function updateGFX(dt)
	if electrics.values.SkyRide_mode == 1 then
		local vel = obj:getVelocity()
		for _, jet in pairs(M.particleEmitters) do
			local n1 = obj:getNodePosition(jet.id1)
			local n2 = obj:getNodePosition(jet.id2)
			local dir = (n2 - n1):normalized()
			obj:addParticleByNodesRelative(jet.id2, jet.id1, -(4 + max(dir:dot(vel)*1.2, 0)), 666, 0.05, 3)
		end
		obj:setVolume(flight_noise, math.min(electrics.values.airspeed/6, 0.75))
		obj:setPitch(flight_noise, 1 + electrics.values.airspeed/3)
		electrics.values.wheelspeed = electrics.values.airspeed
	end
end

M.onPhysicsStep = update
M.updateGFX = updateGFX

local function reset()
	if #M.stabs == 0 then
		M.onPhysicsStep = nop
		M.updateGFX = nop
		return
	end

	obj:stopSFX(flight_noise)

	electrics.values.SkyRide_mode = 0
	M.active = false

	local totalMass, _, _ = extensions.vehicleEditor_nodes.calculateNodesWeight()
	M.totalMass = totalMass
	
	stabs = {}

	aPID:reset()
	pPID:reset()
	rPID:reset()
end

M.onReset = reset

local function init()
	M.stabs = {}

	M.thrusts = {}

	M.particleEmitters = {}
	
	if v.data.wheels then
		for _, thruster in pairs(v.data.wheels) do
			thruster.id1 = thruster.node2
			thruster.id2 = thruster.node1

			table.insert(M.stabs, thruster)

			table.insert(M.particleEmitters, thruster)
		end

		local lookup = {}

		local idt = {}
		
		for ke, va in pairs(M.stabs) do
			if lookup[va.id1] or lookup[va.id2] then
				table.insert(idt, ke)
			end

			lookup[va.id1] = true
			lookup[va.id2] = true
		end

		for ke = #idt, 1, -1 do
			local va = idt[ke]
			table.remove(M.stabs, va)
		end
	end

	local noisemaker = v.data.refNodes[0].cid
	flight_noise = obj:createSFXSource2("art/sound/speed_noise.ogg", "AudioDefaultLoop3D", "flying noise", noisemaker, 0)
	obj:stopSFX(flight_noise)
	transform_noise = obj:createSFXSource2("event:>Vehicle>Hydraulics>Pump_Big", "AudioDefault3D", "Transformation noise", noisemaker, 0)

	enablePhysicsStepHook()

	M.onReset()
end

M.onExtensionLoaded = init
M.onInit = init

M.changeMode = changeMode
M.hoverMode = hoverMode
M.altChange = altChange

return M