--SkyRide Skydros
--Made by: Kanderman
--Do not redistribute

local M = {}

M.skydros = {}

M.torsions = {}

M.onPhysicsStep = nop

electrics.values.skydros_speed = 0

local torsionState = true

local posit = 0

local lastTime = 0

local function clamp(val, min, max) return math.min(math.max(val, min), max) end

local function sign(val) if val > 0 then return 1 elseif val < 0 then return -1 else return 0 end end

local function checkString(val)
    if type(val) == "string" then
        return val
    end
    return ""
end

local function checkNum(val, default)
    return type(val) == 'number' and val or (default or 0)
end

local function setTorsions(b)
    if torsionState == b then
        return
    end
    torsionState = b
    for _, bar in pairs(M.torsions) do
        local sprval = 0
        if b then
            sprval = bar.dspring
        end
        obj:setTorsionbar(bar.bcid, bar.id1, bar.id2, bar.id3, bar.id4, sprval, sprval, bar.damp, bar.damp, checkNum(bar.strength, math.huge), checkNum(bar.deform, math.huge), checkNum(bar.precompressionAngle))
    end
end

local function toggleMode(bool)
    if bool then
        electrics.values.skydros_speed = 1
        setTorsions(false)
    else
        electrics.values.skydros_speed = -1
    end
end

local function update(dt)
    local hydros = M.skydros
    local hcount = #M.skydros

    lastTime = lastTime - dt

    if lastTime <= 0 then
        posit = clamp(posit + electrics.values.skydros_speed, 0, 20)
        if posit == 0 and electrics.values.skydros_speed == -1 then
            setTorsions(true)
        end
        lastTime = 0.0125
    end

    for i = 1, hcount do
        local bee = hydros[i]

        if posit > 0 and lastTime == 0.0125 then
            bee.cmd = bee.mult[posit]
            bee.spml = math.abs(bee.cmd - bee.state)*15
        elseif lastTime == 0.0125 then
            bee.cmd = 1
            bee.spml = math.abs(bee.cmd - bee.state)*15
        end

        if bee.cmd ~= bee.state then
            if bee.cmd < bee.state then
                bee.state = math.max(bee.state - bee.spml*dt, bee.cmd)
            else
                bee.state = math.min(bee.state + bee.spml*dt, bee.cmd)
            end
            bee.fsetRelDisplacement(obj, bee.bcid, bee.state)
        end
    end
end

local function init()
    M.skydros = {}

    M.torsions = {}

    for _, whe in pairs(v.data.wheels) do
        for _, bee in pairs(v.data.beams) do
            if not string.find(checkString(bee.partOrigin), "wheeldata") then
                if whe.node1 == bee.id1 then
                    bee.fIsBroken = obj.beamIsBroken
                    bee.fgetDisplacement = obj.getBeamLength
                    bee.fsetRelDisplacement = obj.setBeamLengthRefRatio
                    bee.fsetRelDeformedDisplacement = obj.setBeamLengthRefDeformedRatio
                    bee.bcid = bee.cid
                    bee.whnd = whe.node2
                    bee.movenode = 1
                    table.insert(M.skydros, bee)
                elseif whe.node1 == bee.id2 then
                    bee.fIsBroken = obj.beamIsBroken
                    bee.fgetDisplacement = obj.getBeamLength
                    bee.fsetRelDisplacement = obj.setBeamLengthRefRatio
                    bee.fsetRelDeformedDisplacement = obj.setBeamLengthRefDeformedRatio
                    bee.bcid = bee.cid
                    bee.whnd = whe.node2
                    bee.movenode = 2
                    table.insert(M.skydros, bee)
                end
            end
        end
        if v.data.torsionbars then
            for _, bar in pairs(v.data.torsionbars) do
                if whe.node1 == bar.id1 or whe.node1 == bar.id4 then
                    bar.bcid = bar.cid
                    bar.dspring = bar.spring
                    table.insert(M.torsions, bar)
                end
            end
        end
    end

    enablePhysicsStepHook()
    M.onReset()
end

local function reset()
    electrics.values.skydros_speed = 0

    posit = 0

    torsionState = true

    if #M.skydros == 0 then
        M.onPhysicsStep = nop
        return
    else
        M.onPhysicsStep = update
    end

    for _,bee in pairs(M.skydros) do
        bee.state = 1 + 1e-28
        bee.cmd = 1

        bee.fsetRelDisplacement(obj, bee.bcid, bee.state)

        local n1, n2, n3

        if bee.movenode == 1 then
            n1 = obj:getNodePositionRelative(bee.id2)
            n2 = obj:getNodePositionRelative(bee.id1)
            n3 = obj:getNodePositionRelative(bee.whnd)
        else
            n1 = obj:getNodePositionRelative(bee.id1)
            n2 = obj:getNodePositionRelative(bee.id2)
            n3 = obj:getNodePositionRelative(bee.whnd)
        end

        if not n3 or not n2 or not n1 then return end

        local ac = n1:distance(n2)
        local ab = n3:distance(n2)

        local side = -1

        if n2.x < 0 then
            side = 1
        end

        local srot = (n3 - n2):normalized():getRotationTo(vec3(0,0,1))

        local frot = quatFromEuler(0, 2.425*side, 0)
 
        bee.mult = {}

        for i = 1, 20 do
            local rot = srot:nlerp(frot, (i - 1)/19)

            local res = n1:distance(vec3(0,0,ab):rotated(rot) + n3)

            bee.mult[i] = res/ac
        end
    end
end

M.onReset = reset

M.onExtensionLoaded = init
M.onInit = init
M.toggleModeActive = toggleMode

return M